// Türklingel mit 100 Melodien M. Baudisch 12/2008
// 20170424, de0508, Uwe
// Anpassung an aktuellen avr gcc version 4.9.2 (GCC) auf meinem System
// 
// kombiniert mit dem Waveform Generator von David Johnson-Davies 
// http://www.technoblogy.com/show?QVN
//
// für ATtiny85, High-Speed-PWM
//
// Compiler: avr-gcc (GCC) 9.2.0


/*
 * Pinout
 * ------
 *
 * Funktion PIN         PIN  Funktion
 *              +--v--+
 *  /RESET  PB5 |1   8| VCC
 *   n.c.   PB3 |2   7| PB2  SCK/PCINT2 /Taster + interner Pull-Up
 *   n.c.   PB4 |3   6| PB1  MISO/OC1A  Lautsprecher
 *          GND |4   5| PB0  MOSI
 *              +-----+
 *
 * Interrupts
 * ----------
 *  PCINT0
 *  TIMER0, Compare Match A
 *
 * Timer
 * -----
 *  Timer 0,  20 kHz
 *  Timer 1, 250 kHz
 */

////////////////////////////////////////
// includes

#include <inttypes.h>       // uint8_t, etc.

#include <avr/io.h>         // Ports & Pins
#include <avr/interrupt.h>  // sei
#include <avr/eeprom.h>     // eeprom Zugriff
#include <avr/pgmspace.h>   // flash Zugriff
#include <avr/sleep.h>      // sleep_mode

#include <util/delay.h>     // _delay_ms

#include "c_makros.h"
#include "lied.h"           // Melodien, Offsets und Tonhoehen


////////////////////////////////////////
// global defines

#define     SPEAKER         (PB1)
#define     TASTE           (PB2)
#define     SOUNDON         set_bit( DDRB, SPEAKER);
#define     SOUNDOFF        clear_bit( DDRB, SPEAKER);

// Zufallszahlengenerator
#define MYRAND_MULTIPLIER  17
#define MYRAND_CONSTANT    37

////////////////////////////////////////
// global types
typedef uint8_t (*function_t)(void);


////////////////////////////////////////
// global variables

volatile register uint16_t soundtime asm("r2");

register uint16_t liedcount    asm("r4");  
register uint16_t liedende     asm("r6");
register uint8_t liednr        asm("r8");
register uint8_t speed         asm("r9");
register uint8_t dauer         asm("r10");
register uint8_t hoehe         asm("r11");
register uint8_t note          asm("r12");

// Initialisierungswert des Zufallsgenerators ist EEPROM Speicherstelle
uint8_t EEMEM seed;
uint8_t g_nLastRandom;

uint16_t    accu;
uint16_t    step;
uint8_t     sine_table[ 256];
function_t  wave_function;


////////////////////////////////////////
// function prototypes
uint8_t square_wave( void);
uint8_t sawtooth_wave( void);
uint8_t triangle_wave( void);
uint8_t sine_wave( void);
uint8_t MyRandGet( void);


////////////////////////////////////////
// helper functions
void calc_sine_table( void)
{
    int16_t x = 0;
    int16_t y = 8180;
    for( int i = 0; i < 256; i++)
    {
        x = x + ( y * 4) / 163;
        y = y - ( x * 4) / 163;
        sine_table[i] = 128 + ( x >> 6);
    }
}


uint8_t MyRandGet( void)
{
  // copy into local variables to avoid 16-bit integer expansion
  uint8_t Multiplier = MYRAND_MULTIPLIER;
  uint8_t Constant = MYRAND_CONSTANT;

  // implicit %256 by 8-bit datatype overflow
  g_nLastRandom = (g_nLastRandom * Multiplier + Constant);

  return g_nLastRandom % ANZ;
}


void setup( void)
{

    set_sleep_mode( SLEEP_MODE_PWR_DOWN);

    // Eingang + Pull-Up
    clear_bit( DDRB, TASTE);
    set_bit( PORTB, TASTE);

    // Pinchange an PB2 konfigurieren
    PCMSK = 1<<PCINT2; 

    // Pinchange Interrupt generell zulassen
    GIMSK = 1<<PCIE; 

    // 64 MHz PLL einschaltun und als Quelle fuer Timer1
    PLLCSR = 1<<PCKE | 1<<PLLE;

    // Timer interrupts aus
    TIMSK = 0;                     

    // Konfiguration Timer/Counter1 fuer PWM-Ausgabe
    // Frequency = 64 MHz / 256 = 250 kHz
    // clear OC1A on compare
    TCCR1 = (1<<PWM1A)|(2<<COM1A0)|(1<<CS10);
    OCR1C = 255;

    // Konfiguration Timer/Counter0 fuer 20kHz interrupt, Sampleausgabe
    TCCR0A = 3<<WGM00;             // Fast PWM
    TCCR0B = 1<<WGM02 | 2<<CS00;   // 1/8 prescale = 1 MHz
    TIMSK  = 1<<OCIE0A;            // Enable compare match, disable overflow
    OCR0A  = 49;                   // Divide by 50 = 20 kHz

    calc_sine_table();

    // Auswahl der Ausgabefunktion
    wave_function = sine_wave;
    //wave_function = triangle_wave;
    //wave_function = sawtooth_wave;
    //wave_function = square_wave;

    // Zufallsgenerator mit Seed initialisieren
    g_nLastRandom = eeprom_read_byte( &seed);
    eeprom_write_byte( &seed, MyRandGet()); //neuen Seed schreiben

    // da Variablen in Registern, explizite Initialisierung noetig!
    liednr      = 0;
    speed       = 0;
    soundtime   = 0;

    // Interrupts global zulassen
    sei();
}


void loop( void)
{
    if( liednr > 0)         // wenn Melodie aktiv
    {
        if( soundtime == 0) // wenn kein Ton oder Pause läuft
        {
            // keine Tonzeitdauer aktiv
            SOUNDOFF;       // Tonwiedergabe aus
            if (!speed)     // wir sind am Liedanfang
            {
                liedcount = eeprom_read_word( &liedofs[ liednr-1]);
                if( liednr == ANZ) 
                {
                    liedende = sizeof( lied);
                }
                else
                {
                    // bei letztem Lied ist Endwert Größe des Arrays
                    // ansonsten 1 Byte vor nächstem Lied
                    liedende = eeprom_read_word( &liedofs[ liednr]); 
                }
                speed = pgm_read_byte( &lied[ liedcount]);
                SOUNDON;
                liedcount++;
            } 
            else                // mitten im Lied
            {
                _delay_ms( 20); // um Toene zu trennen
                // Ton ausgeben
                note  = pgm_read_byte( &lied[ liedcount]); //Note holen
                hoehe = note & 0x1f;    // Tonhöhe ausmaskieren (5 Bit)
                dauer = eeprom_read_byte( &zeit[ note >> 5]); // 5 Bit nach rechts schieben und Zeit holen
                if( hoehe > 0)          // wenn keine Pause 
                {
                    step = eeprom_read_word( &ton[ hoehe-1]);
                    SOUNDON;
                }


                // Bereich dauer: 2..24
                // Bereich speed: 2..5
                // Bereich soundtime: 4..120
                // da jetzt 20k statt 50 -> zusaetzlich Faktor 400

                // da Variable vom Interrupt während der Zuweisung geaendert werden koennte
                cli();
                soundtime = dauer * speed * 400;
                sei();
                
                liedcount++;

                // Lied zu Ende
                if( liedcount > liedende)
                {
                    speed  = 0;
                    liednr = 0;
                }
            }
        } // (!soundtime)
    } // liednr
    else 
    {
        SOUNDOFF; // falls noch Ton an
        sleep_mode();
        liednr = MyRandGet() + 1;
    }
}


uint8_t square_wave( void)
{
    int8_t temp = accu >> 8;
    return( temp >> 7);
}


uint8_t sawtooth_wave( void)
{
    return( accu >> 8);
}


uint8_t triangle_wave( void)
{
    int8_t temp = accu >> 8;
    int8_t mask = accu >> 7;
    temp = temp ^ mask;
    return( temp << 1);
}


uint8_t sine_wave( void)
{
    return( sine_table[ accu >> 8]);
}


// Interruptroutine ohne Inhalt zum Aufwachen aus dem Sleep
EMPTY_INTERRUPT(PCINT0_vect);


// naechstes Sample ausgeben
ISR( TIMER0_COMPA_vect) 
{
    accu = accu + step;
    OCR1A = wave_function();
    if( soundtime > 0)
    {
        soundtime--;
    }
}


int main( void)
{
    setup();

    while( 1)
    {
        loop();
    }
}
